from __future__ import annotations

import attr

import sentaku
from .api import TodoApp
from .pseudorpc import PseudoRpc
from .ux import TodoUX


class TodoAPI(sentaku.ImplementationContext):
    """example description for a simple todo application"""

    @classmethod
    def from_api(cls, api: TodoApp) -> TodoAPI:
        """
        create an application description for the todo app,
        that based on the api can use either tha api or the ux for interaction
        """
        ux = TodoUX(api)

        rpc = PseudoRpc(api)

        return cls.from_instances([api, ux, rpc])

    create_collection = sentaku.ContextualMethod()


@attr.s(auto_attribs=True)
class TodoItem(sentaku.Element):
    """describing a todo list element"""

    parent: TodoCollection
    name: str
    completed = sentaku.ContextualProperty[bool]()


@TodoAPI.external_for(TodoAPI.create_collection, TodoApp)
@TodoAPI.external_for(TodoAPI.create_collection, TodoUX)
def create_todo_collection(self: TodoAPI, name: str) -> TodoCollection:
    self.impl.create_item(name)
    return TodoCollection(self, name=name)


@attr.s
class TodoCollection(sentaku.Collection):
    """domain object describing a todo list"""

    name: str = attr.ib()
    create_item = sentaku.ContextualMethod()
    get_by = sentaku.ContextualMethod()
    clear_completed = sentaku.ContextualMethod()


@TodoAPI.external_for(TodoItem.completed.setter, TodoApp)
@TodoAPI.external_for(TodoItem.completed.setter, TodoUX)
def completed(self: TodoItem, value: bool) -> None:
    col = self.impl.get_by(self.parent.name)
    elem = col.get_by(self.name)
    elem.completed = value


@TodoAPI.external_for(TodoCollection.create_item, TodoApp)
@TodoAPI.external_for(TodoCollection.create_item, TodoUX)
def create_item(self: TodoCollection, name: str) -> TodoItem:
    collection = self.impl.get_by(self.name)
    elem = collection.create_item(name=name)
    assert elem
    return TodoItem(self, name=name)


@TodoAPI.external_for(TodoCollection.get_by, TodoApp)
@TodoAPI.external_for(TodoCollection.get_by, TodoUX)
def get_by(self: TodoCollection, name: str) -> TodoItem | None:
    collection = self.impl.get_by(self.name)
    elem = collection.get_by(name)
    if elem is not None:
        return TodoItem(self, name=name)
    else:
        return None


@TodoAPI.external_for(TodoCollection.clear_completed, TodoUX)
@TodoAPI.external_for(TodoCollection.clear_completed, TodoApp)
def clear_completed(self: TodoCollection) -> None:
    collection = self.impl.get_by(self.name)
    collection.clear_completed()


@TodoAPI.external_for(TodoItem.completed.setter, PseudoRpc)
def todo_item_set_item_completion(item: TodoItem, value: bool) -> None:
    if value:
        item.impl.complete_item(item.parent.name, item.name)
    else:
        raise NotImplementedError("rpc cant undo completion")


@TodoAPI.external_for(TodoCollection.create_item, PseudoRpc)
def create_todo_item(coll: TodoCollection, name: str) -> TodoItem:
    coll.impl.make_item(coll.name, name)
    return TodoItem(coll, name=name)


@TodoAPI.external_for(TodoCollection.get_by, PseudoRpc)
def get_by_rpc(self: TodoCollection, name: str) -> TodoItem | None:
    if self.impl.has_item(self.name, name):
        return TodoItem(self, name=name)
    else:
        return None


@TodoAPI.external_for(TodoCollection.clear_completed, PseudoRpc)
def clear_completed_rpc(coll: TodoCollection) -> None:
    coll.impl.clear_completed(coll.name)


@TodoAPI.external_for(TodoAPI.create_collection, PseudoRpc)
def create_collection_rpc(api: TodoAPI, name: str) -> TodoCollection:
    api.impl.make_collection(name)
    return TodoCollection(api, name=name)



"""
the implementations_stack manages implementation picking
and fallback preferences

based on the contexts pushed/poped from the stack it will choose
context roots and help picking implementations
"""
from __future__ import annotations

from contextlib import contextmanager
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import NamedTuple
from typing import NoReturn
from typing import Sequence
from typing import Union

from typing_extensions import TypeAlias

LIMIT = 20


class ImplementationChoice(NamedTuple):
    key: object
    value: Any


class Chooser(NamedTuple):
    elements: tuple[object, ...]
    previous: Chooser | NullChooser
    frozen: bool

    @classmethod
    def make(
        cls, previous: Chooser | NullChooser, elements: tuple[object, ...], frozen: bool
    ) -> Chooser:
        if previous.frozen:
            raise RuntimeError(
                "further nesting of implementation choice has been disabled"
            )
        if frozen:
            assert len(elements) == 1

        return cls(elements, previous, frozen)

    def choose(self, choose_from: Mapping[object, Any]) -> ImplementationChoice:
        """given a mapping of implementations
        choose one based on the current settings
        returns a key value pair
        """

        for choice in self.elements:
            if choice in choose_from:
                return ImplementationChoice(choice, choose_from[choice])
        raise LookupError(self.elements, set(choose_from.keys()))


class NullChooser:
    frozen = False
    elements = ()

    def choose(self, *_: object, **__: object) -> NoReturn:
        raise LookupError("No choice possible without valid context")


CHOOSER: TypeAlias = Union[NullChooser, Chooser]


def chain(chooser: CHOOSER) -> list[Chooser]:
    elements: list[Chooser] = []
    while not isinstance(chooser, NullChooser):
        elements.append(chooser)
        chooser = chooser.previous
    elements.reverse()
    return elements


class ChooserStack:
    current: Chooser | NullChooser

    def __init__(
        self, default_elements: CHOOSER | ChooserStack | Sequence[object] | None = None
    ) -> None:
        self.current = NullChooser()

        if default_elements is not None:
            if isinstance(default_elements, (Chooser, NullChooser)):
                self.current = default_elements
            elif isinstance(default_elements, ChooserStack):
                self.current = default_elements.current
            else:
                self.current = Chooser.make(
                    elements=tuple(default_elements),
                    previous=NullChooser(),
                    frozen=False,
                )

    def __repr__(self) -> str:
        return f"<ICS {chain(self.current)}>"

    def choose(self, choose_from: Mapping[object, Any]) -> ImplementationChoice:
        """given a mapping of implementations
        choose one based on the current settings
        returns a key value pair
        """
        return self.current.choose(choose_from)

    @contextmanager
    def pushed(
        self, elements: Sequence[object], frozen: bool = False
    ) -> Iterator[None]:
        previous = self.current
        self.current = Chooser.make(previous, tuple(elements), frozen)
        try:
            if len(chain(self.current)) > LIMIT:
                raise OverflowError("stack depth exceeded")
            yield
        finally:
            self.current = previous



from __future__ import annotations

import contextlib
import warnings
from typing import Any
from typing import Callable
from typing import cast
from typing import ClassVar
from typing import Generic
from typing import Iterator
from typing import Mapping
from typing import overload
from typing import Protocol
from typing import Sequence
from typing import TypeVar

import attr
from typing_extensions import Self

from .chooser import ChooserStack
from .chooser import ImplementationChoice


class HasContext(Protocol):
    @property
    def context(self) -> ImplementationContext:
        ...


METHOD_DATA_KEY = "sentaku_method_data"

F = TypeVar("F")
T = TypeVar("T")


@contextlib.contextmanager
def _use_maybe_strict(ctx: ImplementationContext, implt: object) -> Iterator[None]:
    with ctx.use(implt, frozen=ctx.strict_calls):
        yield


class ImplementationContext:
    """maintains a mapping
    of :ref:`implementation-identification` to implementations,
    as well as the list of currently available Implementations
    in the order of precedence.
    """

    __registrations: ClassVar[dict[object, dict[object, Any]]] = {}
    __combined_registrations: ClassVar[Mapping[object, Mapping[object, Any]]]

    def __init_subclass__(cls) -> None:
        cls.__registrations = {}

    implementations: dict[object, object]
    implementation_chooser: ChooserStack
    strict_calls: bool

    def __init__(
        self,
        implementations: dict[object, object],
        implementation_chooser: ChooserStack,
        strict_calls: bool = False,
    ):
        self.implementations = implementations
        self.implementation_chooser = implementation_chooser
        self.strict_calls = strict_calls

    @classmethod
    def external_for(cls, method: object, implementation: object) -> Callable[[F], F]:
        if isinstance(method, ContextualProperty):
            raise TypeError(
                f"{method} cannot be registered, "
                "use its .getter/setter attributes for registration"
            )

        def register(func: F) -> F:
            if method in cls.__registrations:
                registry = cls.__registrations[method]
            else:
                registry = cls.__registrations[method] = {}

            if implementation in registry:
                raise ValueError("conflict", implementation)

            registry[implementation] = func
            return func

        return register

    @classmethod
    def with_default_choices(
        cls,
        implementations: dict[object, object],
        default_choices: Sequence[object],
        **kw: Any,
    ) -> Self:
        return cls(
            implementations=implementations,
            implementation_chooser=ChooserStack(default_choices),
            **kw,
        )

    @property
    def impl(self) -> Any:
        """the currently active implementation"""
        return self.implementation_chooser.choose(self.implementations).value

    def _get_implementation_for(self, key: object) -> ImplementationChoice:
        try:
            implementation_set = self.__combined_registrations[key]
        except (AttributeError, KeyError):
            combined: dict[object, dict[object, object]] = {}
            context_cls: type[ImplementationContext]
            for context_cls in reversed(type(self).__mro__):
                if context_cls is object:
                    continue
                for method, implementations in context_cls.__registrations.items():
                    if method in combined:
                        combined[method].update(implementations)
                    else:
                        combined[method] = implementations.copy()
            type(self).__combined_registrations = combined
            implementation_set = self.__combined_registrations[key]
        return self.implementation_chooser.choose(implementation_set)

    @classmethod
    def from_instances(cls, instances: Sequence[object], **kw: Any) -> Self:
        """utility to create the context

        by passing an ordered list of instances
        and turning them into implementations and the default choices
        """
        return cls.with_default_choices(
            implementations={type(x): x for x in instances},
            default_choices=[type(x) for x in instances],
            **kw,
        )

    @property
    def context(self) -> Self:
        """alias for consistence with elements"""
        return self

    root = context

    @contextlib.contextmanager
    def use(self, *implementation_types: object, frozen: bool = False) -> Iterator[Any]:
        """contextmanager for controlling
        the currently active/usable implementations and
        their order of precedence

        :param implementation_types:
            the implementations available within the context
        :param bool frozen: if True prevent further nesting
        """

        with self.implementation_chooser.pushed(implementation_types, frozen=frozen):
            yield self.impl


@attr.s
class _ImplementationBindingMethod:
    """bound method equivalent for  :class:`ImplementationCooser`

    when called it

    * looks up the implementation
    * freezes the context
    * calls the actual implementation
    """

    instance: object = attr.ib()
    selector: object = attr.ib()

    def __call__(self, *k: Any, **kw: Any) -> Any:
        ctx = cast(HasContext, self.instance).context
        choice, implementation = ctx._get_implementation_for(self.selector)
        bound_method = implementation.__get__(self.instance, type(self.instance))
        with _use_maybe_strict(ctx, choice):
            return bound_method(*k, **kw)


class ContextualMethod:
    """
    descriptor for implementing context-sensitive methods
    and registration of their implementations


    .. code:: python

        class Example(Element):
            action = ContextualMethod()
            @action.implemented_for("db")
            def action(self):
                pass

           @action.implemented_for("test")
           def action(self):
               pass
    """

    def __set_name__(self, owner: type[HasContext], name: str) -> None:
        pass

    # todo - turn into attrs class once attribute anchoring is implemented
    def __repr__(self) -> str:
        return "<ContextualMethod>"

    @overload
    def __get__(self, instance: None, owner: type[object]) -> ContextualMethod:
        ...

    @overload
    def __get__(
        self, instance: object, owner: type[object]
    ) -> _ImplementationBindingMethod:
        ...

    def __get__(
        self, instance: object | None, owner: object
    ) -> ContextualMethod | _ImplementationBindingMethod:
        if instance is None:
            return self
        return _ImplementationBindingMethod(instance=instance, selector=self)


class ContextualProperty(Generic[T]):
    name: str

    def __set_name__(self, owner: object, name: str) -> None:
        self.name = name

    def getter(self, func: Callable[[object], T]) -> Self:
        warnings.warn(f"{self.getter} is not yet registering {func}")
        return self

    def setter(self, func: Callable[[object, T], None]) -> Self:
        warnings.warn(f"{self.setter} is not yet registering {func}")
        return self

    def __set__(self, instance: HasContext, value: T) -> None:
        ctx = instance.context
        choice, implementation = ctx._get_implementation_for(self.setter)

        bound_method = implementation.__get__(instance, type(instance))
        with _use_maybe_strict(ctx, choice):
            bound_method(value)

    @overload
    def __get__(self, instance: None, owner: object) -> Self:
        ...

    @overload
    def __get__(self, instance: HasContext, owner: object) -> Any:
        ...

    def __get__(self, instance: HasContext | None, owner: object) -> Any | Self:
        if instance is None:
            return self

        ctx = instance.context
        choice, implementation = ctx._get_implementation_for(self.getter)

        bound_method = implementation.__get__(instance, type(instance))
        with _use_maybe_strict(ctx, choice):
            return bound_method()



from typing import Any

import attr
from typing_extensions import assert_type

from sentaku import ContextualMethod
from sentaku import ContextualProperty
from sentaku import Element


@attr.s
class Fun(Element):
    lol: str = attr.ib()
    prop = ContextualProperty[Any]()
    method = ContextualMethod()


assert_type(Fun.prop, ContextualProperty[Any])
assert_type(Fun.method, ContextualMethod)



from __future__ import annotations

from typing import cast
from typing import Union

import pytest

from sentaku import Element
from sentaku.context import ContextualMethod
from sentaku.context import ContextualProperty
from sentaku.context import ImplementationContext


def test_from_instances() -> None:
    ctx: ImplementationContext = ImplementationContext.from_instances([1, "a"])
    assert ctx.implementation_chooser.current.elements == (int, str)
    assert ctx.implementations == {int: 1, str: "a"}


class LocalContext(ImplementationContext):
    pass


class LocalElement(Element):
    method = ContextualMethod()
    prop = ContextualProperty[Union[int, str]]()


@LocalContext.external_for(LocalElement.method, int)
@LocalContext.external_for(LocalElement.method, str)
@LocalContext.external_for(LocalElement.prop.setter, int)
@LocalContext.external_for(LocalElement.prop.setter, str)
@LocalContext.external_for(LocalElement.prop.getter, int)
@LocalContext.external_for(LocalElement.prop.getter, str)
def method_standin(self: LocalElement, value: int | str | None = None) -> int:
    assert (
        self.context.implementation_chooser.current.frozen == self.context.strict_calls
    )
    return 1


@pytest.fixture(
    params=[
        pytest.param(True, id="nest=strict"),
        pytest.param(False, id="nest=lenient"),
    ]
)
def ctx(request: pytest.FixtureRequest) -> LocalContext:
    return LocalContext.from_instances([1, "a"], strict_calls=request.param)


@pytest.fixture
def elem(ctx: LocalContext) -> LocalElement:
    return LocalElement(parent=ctx)


@pytest.fixture(params=(int, str))
def impl(request: pytest.FixtureRequest) -> type:
    return cast(type, request.param)


def test_property(ctx: LocalContext, elem: LocalElement, impl: type) -> None:
    with ctx.use(impl):
        assert elem.prop == 1


def test_set_property(ctx: LocalContext, elem: LocalElement, impl: type) -> None:
    with ctx.use(impl):
        elem.prop = 1
        assert elem.prop == 1


def test_register_prop_fails() -> None:
    with pytest.raises(TypeError):
        LocalContext.external_for(LocalElement.prop, int)


def test_method(ctx: LocalContext, elem: LocalElement, impl: type) -> None:
    with ctx.use(impl):
        elem.method()


def test_delayed_register_method(ctx: LocalContext, elem: LocalElement, impl: type) -> None:
    with ctx.use(impl):
        # Call a registered contextual method
        elem.method()

        # Register a new contextual method
        class NewLocalElement(LocalElement):
            new_method = ContextualMethod()

        @LocalContext.external_for(NewLocalElement.new_method, int)
        @LocalContext.external_for(NewLocalElement.new_method, str)
        def new_method_standin(self: NewLocalElement, value: int | str | None = None) -> int:
            assert (
                self.context.implementation_chooser.current.frozen == self.context.strict_calls
            )
            return 1

        # Call both registered methods
        new_elem = NewLocalElement(parent=ctx)
        new_elem.method()
        new_elem.new_method()
        elem.method()




import pytest

from sentaku.chooser import ChooserStack


@pytest.fixture
def chooser() -> ChooserStack:
    return ChooserStack(None)


def test_empty(chooser: ChooserStack) -> None:
    with pytest.raises(LookupError):
        chooser.choose({3: 3})


def test_nonempty(chooser: ChooserStack) -> None:
    with chooser.pushed((1,)):
        assert chooser.current.elements == (1,)


def test_freezing(chooser: ChooserStack) -> None:
    with chooser.pushed([1], frozen=True):
        with pytest.raises(RuntimeError):
            with chooser.pushed([1]):
                pass


def test_overflow(chooser: ChooserStack) -> None:
    def nest(n: int = 21) -> None:
        if n:
            with chooser.pushed([1]):
                nest(n=n - 1)

    with pytest.raises(OverflowError):
        nest()


def test_choose(chooser: ChooserStack) -> None:
    with chooser.pushed([1, 2]):
        res = chooser.choose({1: "a"})
    assert res == (1, "a")


def test_choose_missing(chooser: ChooserStack) -> None:
    with chooser.pushed([1, 2]):
        with pytest.raises(LookupError):
            chooser.choose({3: 3})



from .context import ContextualMethod
from .context import ContextualProperty
from .context import ImplementationContext
from .modeling import Collection
from .modeling import Element


__all__ = [
    "Element",
    "Collection",
    "ContextualMethod",
    "ContextualProperty",
    "ImplementationContext",
]



"""
a hideous rpc interface used to demonstrate partial implementation
"""
from __future__ import annotations

from typing import Any

import attr


@attr.s
class PseudoRpc:
    """a hideous implementation"""

    _backend: Any = attr.ib(repr=False)

    def make_collection(self, collection: str) -> None:
        """creates a todo list"""
        self._backend.create_item(collection)

    def make_item(self, collection: str, name: str) -> None:
        """creates a todo list element"""
        self._backend.get_by(collection).create_item(name)

    def has_item(self, collection: str, name: str) -> bool:
        """checks for the existence of a todo list element"""
        return self._backend.get_by(collection).get_by(name) is not None

    def complete_item(self, collection: str, name: str) -> None:
        """marks a todo list element as completed"""
        self._backend.get_by(collection).get_by(name).completed = True

    def clear_completed(self, collection: str) -> None:
        """clears the completed elements of a todo list"""
        self._backend.get_by(collection).clear_completed()



from __future__ import annotations

from typing import Any

import attrs

from . import context as _ic


class ElementMixin:
    parent: ElementMixin | _ic.HasContext

    @property
    def context(self) -> _ic.ImplementationContext:
        """the context this element belongs to"""
        return self.parent.context

    @property
    def impl(self) -> Any:
        """shortcut to get the currently active application implementation"""
        return self.context.impl


@attrs.define(slots=False)
class Element(ElementMixin):
    """Base class for all application elements

    :param parent:
        controlling object of the element which is either a context
        or a surrounding element
    :type parent: :py:class:`Element` or :py:class:`ImplementationContext`
    """

    parent: Element | _ic.HasContext = attrs.field(repr=False)


@attrs.define(slots=False)
class Collection(Element):
    """base class for collections in the application

    :todo: generic helpers for querying
    """



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Iterator
from typing import Protocol
from typing import TypeVar

import attr

T = TypeVar("T")


class HasName(Protocol):
    name: str


@attr.s
class TodoElement:
    """
    Element of a todo list
    """

    name: str = attr.ib()
    completed: bool = attr.ib(default=False)


def get_by(self: Iterator[HasName], name: str) -> HasName | None:
    """get element by name"""
    for item in self:
        if item.name == name:
            return item
    else:
        return None


def create_by_name(cls: type[T], collection_name: str) -> Callable[[Any, str], T]:
    def create_item(self: Any, name: str) -> T:
        assert self.get_by(name) is None
        item = cls(name=name)  # type: ignore  [call-arg]
        getattr(self, collection_name).append(item)
        return item

    create_item.__doc__ = f"create a new named {cls} item"

    return create_item


@attr.s
class TodoList:
    """a named todolist"""

    name: str = attr.ib()
    items: list[TodoElement] = attr.ib(default=attr.Factory(list), converter=list)

    def __iter__(self) -> Iterator[TodoElement]:
        return iter(self.items)

    get_by = get_by
    create_item = create_by_name(TodoElement, "items")

    def clear_completed(self) -> None:
        """
        removes completed elements
        """
        self.items = [i for i in self.items if not i.completed]


@attr.s
class TodoApp:
    """
    A Basic Todo List Storage

    """

    collections: list[TodoList] = attr.ib(default=attr.Factory(list), converter=list)

    def __iter__(self) -> Iterator[TodoList]:
        return iter(self.collections)

    def __repr__(self) -> str:
        return f"<TodoApp {sorted(x.name for x in self)!r}>"

    get_by = get_by
    create_item = create_by_name(TodoList, "collections")
